/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2011 Intel Corporation. All rights reserved.
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

/*
 * In Memory Fuzzer - a tool to fuzz a function in memory, randomizing
 * the first input argument value. This could be the bae for a 
 * checkpointing approach to accelerate fuzzing
 */

#include "pin.H"
#include <iostream>
#include <fstream>
#include <time.h>
#include <stdlib.h>

/*
 * Globals
 */
CONTEXT SavedCtxt;

std::ofstream out;						

KNOB<string> KnobOutputFile(KNOB_MODE_WRITEONCE,  "pintool",
    "o", "InMemFuzz.out", "specify file name for the output file");

KNOB<string> KnobCheckpointFunc(KNOB_MODE_WRITEONCE,  "pintool",
    "c", "AfterInit", "specify the name of the function to hang the checkpoint on");

/*
 * Analysis Routines
 */
VOID SaveCheckPoint(CONTEXT * CurCtxt)
{
	//actually we only need this function once so it would be smart 
	//to handle this as conditional instrumentation

	//important! you can't copy contexts without using this func
	PIN_SaveContext(CurCtxt, &SavedCtxt);
}


VOID RestoreCheckPoint(CONTEXT * CurCtxt)
{
	out << time(NULL) << " Restoring context" << endl;

	//you probably want to restore stack values and global data structures here
	
	//we should load the new inputs here
	ADDRINT ESP = PIN_GetContextReg( &SavedCtxt, REG_ESP);
	//ESP points to top of stack after call and before execution of instruction
	//therefore the first param is at ESP + 4 (IA32 calling convention only!)
	ESP += 4;
	//write new random value to memory address
	(*((int *)ESP)) = rand() % 10000000; //limit rand range 
	
	//loads the context and never returns
	PIN_ExecuteAt(&SavedCtxt);
}


/*
 * Instrumentation Callbacks
 */
VOID Image(IMG img, VOID *v)
{
	RTN rtn = RTN_FindByName( img, KnobCheckpointFunc.Value().c_str() );
	if ( RTN_Valid(rtn) )
	{
		RTN_Open(rtn);
        RTN_InsertCall(rtn, IPOINT_BEFORE, (AFUNPTR)SaveCheckPoint,
               IARG_CONST_CONTEXT, IARG_END); //CONST_CONTEXT is 4x faster
		RTN_Close(rtn);
	}

	rtn = RTN_FindByName( img, "main" );
	if ( RTN_Valid(rtn) )
	{
		RTN_Open(rtn);
        RTN_InsertCall(rtn, IPOINT_AFTER, (AFUNPTR)RestoreCheckPoint, IARG_END);
		RTN_Close(rtn);
	}
}


VOID Fini(INT32 code, VOID *v)
{
    out.close();
}

/* ===================================================================== */
/* Main                                                                  */
/* ===================================================================== */

int main(int argc, char *argv[])
{
    // Initialize pin & symbol manager
    PIN_InitSymbols();
    PIN_Init(argc,argv);

    // Initialize RNG
    srand( time(NULL) );

	// Initialize logfile
    string fileName = KnobOutputFile.Value();
    out.open(fileName.c_str());
    out << hex;

    // Register Image to be called to instrument functions.
    IMG_AddInstrumentFunction(Image, 0);
	PIN_AddFiniFunction(Fini, 0);

    PIN_StartProgram(); // Never returns
    
    return 0;
}
